import dataclasses
import json
import logging
import string
import struct
from typing import Dict, Iterator, Tuple

import click
import prettytable
from elftools.elf.constants import SH_FLAGS
from elftools.elf.elffile import ELFFile

_LOG = logging.getLogger(__name__)


@dataclasses.dataclass
class LogFormatInfo:
    filename: str
    line: int
    n_args: int
    fmt: str


LOG_FMT_SECTION_ELF_HEADER_MAGIC = 0x66474F4C  # 'LOGf'
LOG_FMT_SECTION_ELF_VERSION = 1
LOG_FMT_HDR_LEN = 8


@dataclasses.dataclass
class Header:
    magic: int
    version: int

    @classmethod
    def parse(cls, data: bytes) -> "Header":
        # Maps to sMemfaultLogFmtElfSectionHeader from compact_log_helpers.h
        magic, version = struct.unpack("<IBxxx", data[0:LOG_FMT_HDR_LEN])
        return cls(magic & 0xFFFFFFFF, version)


class LogFormatElfSectionParser:
    def find_section(self, elf: ELFFile) -> Tuple[Header, bytes]:
        for section in elf.iter_sections():
            if section.name != "log_fmt":
                continue

            sh_flags = section["sh_flags"]

            if sh_flags & SH_FLAGS.SHF_ALLOC != 0:
                raise ValueError(
                    "The 'log_fmt' section should not be allocated. Mark it as (INFO) in the ELF"
                )

            data = section.data()
            header = Header.parse(data)

            if header.magic != LOG_FMT_SECTION_ELF_HEADER_MAGIC:
                raise ValueError(f"Invalid 'log_fmt' section magic: 0x{header.magic:02x}")

            return header, data[LOG_FMT_HDR_LEN:]

        raise ValueError("Could not locate 'log_fmt' section")

    def parse(self, section: Tuple[Header, bytes]) -> Dict[int, LogFormatInfo]:
        header, body = section

        if header.version != LOG_FMT_SECTION_ELF_VERSION:
            raise ValueError(f"Unsupported 'log_fmt' version: {header.version}")

        return dict(self._parse(body))

    def _parse(self, body: bytes) -> Iterator[Tuple[int, LogFormatInfo]]:
        first_error = None
        log_fmt_id = LOG_FMT_HDR_LEN

        # Note: Each log_fmt entry should be separated by a NUL character because we are parsing a list
        # of strings generated by the C Compiler
        sep = b"\0"
        for part in body.split(sep):
            try:
                if part:
                    entry = self._parse_entry(part)
                    yield log_fmt_id, entry
            except ValueError as exc:
                first_error = exc, part
            log_fmt_id += len(part) + len(sep)

        # avoid spamming log output
        if first_error:
            _LOG.warning("Failed to parse: %s", first_error)

    def _parse_entry(self, entry: bytes) -> LogFormatInfo:
        # The format of a V1 entry: (n_args;__FILE__;__LINE__;fmt_str)
        # See MEMFAULT_LOG_FMT_ELF_SECTION_ENTRY macro in compact_log_helpers.h
        fields = entry.split(b";", maxsplit=3)
        n_args_bytes, filename, line, log_fmt = fields
        try:
            n_args = int(n_args_bytes)
        except ValueError:
            # Empirically, certain compilers (such as the TI ARM Compiler) may emit
            # junk (padding?) bytes between strings. Therefore, we strip any leading
            # non-decimal characters from n_args to work around the problem.
            digits = set(string.digits.encode())
            n_args_bytes = bytes(c for c in n_args_bytes if c in digits)
            n_args = int(n_args_bytes)

        return LogFormatInfo(
            filename=filename.decode(),
            line=int(line),
            n_args=int(n_args),
            fmt=log_fmt.decode(),
        )

    @classmethod
    def get_mapping_from_elf_file(cls, filepath: str) -> Dict[int, LogFormatInfo]:
        """Convenience method to loading mapping from an ELF file path"""
        with open(filepath, "rb") as f:
            elf = ELFFile(f)
            parser = cls()
            section = parser.find_section(elf)
            return parser.parse(section)


@click.command()
@click.argument("elf", type=click.Path(exists=True))
@click.option(
    "-i",
    "--log_id",
    default=None,
    help="log_id to filter on. If not provided entire list is dumped",
)
@click.option(
    "--json-output",
    is_flag=True,
    show_default=True,
    default=False,
    help="Ouput in JSON format instead of table",
)
def main(elf, log_id, json_output):
    """
    Given an ELF, looks for the log_fmt section & dumps all formatters found

    Example Usage:

    \b
    $ python format.py my_build.elf
        Found 3 log formats
        +--------+-----------+------+----------+----------+
        | Log Id | File Name | Line | Num Args |  Format  |
        +--------+-----------+------+----------+----------+
        |   8    |   main.c  | 151  |    1     | Hello %d |
        |                      ...                        |
        +--------+-----------+------+----------+----------+

    """
    mappings = LogFormatElfSectionParser.get_mapping_from_elf_file(elf)

    if json_output:
        output = {}
        for k, v in mappings.items():
            output[k] = dataclasses.asdict(v)

        # sort by log_id; should be sorted by default but just in case
        output = {k: v for k, v in sorted(output.items(), key=lambda item: item[0])}
        click.echo(json.dumps(output))
        return
    else:
        pt = prettytable.PrettyTable(["Log Id", "File Name", "Line", "Num Args", "Format"])

        log_ids = list(mappings.keys())
        log_ids.sort()
        print(f"Found {len(log_ids)} log formats")

        if log_id is not None:
            log_ids = [n for n in log_ids if n == int(log_id)]

        for k in log_ids:
            v = mappings[k]
            pt.add_row([k, v.filename, v.line, v.n_args, v.fmt])

        print(pt)


if __name__ == "__main__":
    main()



import functools
import logging
import os
import platform
import struct
from contextlib import contextmanager
from enum import IntEnum
from typing import Any, List, Tuple

import cbor2
import click
from wasmer import Instance, Module, Store, engine

LOG = logging.getLogger(__name__)

FILE_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)))
WASM_EXE_DIR = os.path.join(FILE_DIR, "decode")
WASM_BYTECODE = os.path.join(FILE_DIR, "decode", "decode.wasm")


@functools.lru_cache(maxsize=1)
def load_wasm_program_serialized():
    machine = platform.machine().lower()  # i.e x86_64
    system = platform.system().lower()  # i.e linux
    wasm_exe = f"decode_{machine}-{system}.so"
    native_exe_path = os.path.join(WASM_EXE_DIR, wasm_exe)
    if os.path.exists(native_exe_path):
        store = Store(engine.Native())
        LOG.info("Using Native WASM engine")
        with open(native_exe_path, "rb") as f:
            return (store, f.read())
    else:
        # We don't have a pre-compiled runtime so compile on the fly
        # Lazy load import so we don't have to pull in dep for pre-compiled use case
        #
        # WARNING: We have seen random compiler hangs when this path is hit from certain
        # environments (i.e forked python child processes). If a hang is observed, first
        # try building a native binary using the native_engine.py script.
        LOG.info("Running a JIT WASM compilation")
        from wasmer_compiler_cranelift import Compiler

        store = Store(engine.JIT(Compiler))
        with open(WASM_BYTECODE, "rb") as f:
            return store, Module(store, f.read()).serialize()


class CompactLogPromotionType(IntEnum):
    INT32 = 0
    INT64 = 1
    DOUBLE = 2
    STRING = 3


class CompactLogDecoder:
    def __init__(self):
        store, serialized = load_wasm_program_serialized()
        self.wasm_instance = Instance(Module.deserialize(store, serialized))
        self.allocations = []

    @staticmethod
    def unpack_compact_log(
        input_args: List,
    ) -> Tuple[int, List[Tuple[CompactLogPromotionType, object]]]:
        """
        Unpacks CBOR encoded "compact logs" into the log_id & a va_arg array

        where,
         - log_id The id used to recover the C fmt string from an ELF
         - va_args A list of [ argument type, argument value ]

        More details about the encoding scheme in:
          sdk/embedded/components/log/src/memfault_compact_log_serializer.c
        """
        args = iter(input_args)
        log_fmt_id = next(args)
        arg_list: List[Tuple[CompactLogPromotionType, Any]] = []

        for arg in args:
            if isinstance(arg, int):
                arg_list.append((CompactLogPromotionType.INT32, arg))
            elif isinstance(arg, list):
                arg_list.append((CompactLogPromotionType.INT64, arg[0]))
            elif isinstance(arg, float):
                arg_list.append((CompactLogPromotionType.DOUBLE, arg))
            elif isinstance(arg, str):
                arg_list.append((CompactLogPromotionType.STRING, arg))
            else:
                raise ValueError("Unknown arg type {}".format(arg))

        return log_fmt_id, arg_list

    @contextmanager
    def _alloc_tracking(self):
        try:
            yield
        finally:
            for ptr in self.allocations:
                self.wasm_instance.exports.log_free(ptr)
            self.allocations = []

    def _zmalloc(self, length: int):
        pointer = self.wasm_instance.exports.log_zmalloc(length)
        self.allocations.append(pointer)
        return pointer

    def strdup(self, in_str: str):
        """
        Give a Python string, build a C NUL termintated string in the WASM runtime
        """
        # We are building a null terminated c-string
        pointer = self._zmalloc(len(in_str) + 1)

        # Copy python string into wasm environment
        memory = self.wasm_instance.exports.memory.uint8_view(pointer)
        for idx, c in enumerate(in_str.encode()):
            memory[idx] = c

        # return reference
        return pointer

    def build_va_args(self, va_args: List[Tuple[CompactLogPromotionType, object]]):
        """
        Converts va_args as a Python List to a C-lang suitable format in the WASM runtime

        Note: The "..." (va_args) in C are just a pointer to a bytearray of arguments. The
        "fmt" is used for recovering the actual args. Arguments will always be 4 or 8 bytes
        and double-word args need to be double-word aligned.
        """
        curr_offset = 0
        c_va_args = bytearray()

        for arg_type, arg in va_args:
            # First check if it's a double-word and we need to pad
            # out to aligned boundary
            if arg_type in (CompactLogPromotionType.INT64, CompactLogPromotionType.DOUBLE):
                if curr_offset % 8 != 0:
                    # Pad out to double-word aligned boundary
                    c_va_args.extend(struct.pack("<I", 0))
                    curr_offset += 4

            if arg_type == CompactLogPromotionType.INT32:
                assert isinstance(arg, int)
                encoded_arg = struct.pack("<I", arg & 0xFFFFFFFF)
            elif arg_type == CompactLogPromotionType.INT64:
                assert isinstance(arg, int)
                encoded_arg = struct.pack("<Q", arg)
            elif arg_type == CompactLogPromotionType.DOUBLE:
                assert isinstance(arg, (int, float))
                encoded_arg = struct.pack("d", arg)
            elif arg_type == CompactLogPromotionType.STRING:
                assert isinstance(arg, str)
                str_ptr = self.strdup(arg)
                encoded_arg = struct.pack("<I", int(str_ptr))
            else:
                raise ValueError("Unexpected Arg {} {}".format(arg_type, arg))

            c_va_args.extend(encoded_arg)
            curr_offset += len(encoded_arg)

        # We've built what the memory layout for va_args should look like
        # Now let's copy it into the C env
        va_args_ptr = self._zmalloc(curr_offset)
        args_view = self.wasm_instance.exports.memory.uint8_view(va_args_ptr)
        for idx, b in enumerate(c_va_args):
            args_view[idx] = b

        return va_args_ptr

    def decode(self, c_fmt: str, va_args: List, *, max_result_len=512) -> bytes:
        """
        Converts compact log to a full byte string given a C fmt string & arg list

        Args:
          c_fmt: A format string suitable for libc's printf family of functions
          va_args: Argument output from unpack_compact_log
          max_result_len: Maximum length string to generate before truncating decode
        """
        with self._alloc_tracking():
            result_ptr = self._zmalloc(max_result_len)

            c_fmt_ptr = self.strdup(c_fmt)

            va_args_ptr = self.build_va_args(va_args)

            self.wasm_instance.exports.log_vsnprintf(
                result_ptr, max_result_len, c_fmt_ptr, va_args_ptr
            )

            result = self.wasm_instance.exports.memory.uint8_view(result_ptr)
            result_str = bytes(result[0:max_result_len]).split(b"\x00")[0]
        return result_str

    @classmethod
    def from_cbor_array(cls, fmt_str, compact_log: bytes):
        """
        Helper for 1-off decoding a compact log.
        """
        args = cbor2.loads(compact_log)
        if not isinstance(args, list):
            raise ValueError("Type is {}. Expected <class 'list'>".format(type(args)))

        _, va_args = cls.unpack_compact_log(args)
        compact_log_decoder = cls()
        return compact_log_decoder.decode(fmt_str, va_args)

    @classmethod
    def from_arg_list(cls, fmt_str, args: list):
        """
        Helper for 1-off decoding a format log that has been deserialized from cbor
        """
        _, va_args = cls.unpack_compact_log(args)
        compact_log_decoder = cls()
        return compact_log_decoder.decode(fmt_str, va_args)


@click.command()
@click.argument("fmt")
@click.argument("compact-log")
def main(fmt, compact_log):
    """
    Given a fmt string and hex string encoded compact log, recover original
    string

    Example Usage:

    \b
    $ python compact.py "An Integer Format String: %d" 820A0B
        Log Recovered:
        An Integer Format String: 11

    """
    compact_log_as_bytes = bytes.fromhex(compact_log)
    log = CompactLogDecoder.from_cbor_array(fmt, compact_log_as_bytes)
    click.secho("Log Recovered:")
    click.secho(log.decode(), fg="green")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['mflt_compact_log']

package_data = \
{'': ['*'],
 'mflt_compact_log': ['decode/decode.wasm',
                      'decode/decode.wasm',
                      'decode/decode_x86_64-linux.so',
                      'decode/decode_x86_64-linux.so']}

install_requires = \
['cbor2>=5,<6',
 'click>=8.0,<9.0',
 'prettytable==0.7.2',
 'pyelftools>=0.28.0,<0.29.0']

extras_require = \
{':python_version >= "3.10" and python_version < "3.11"': ['wasmer-py310==1.0.0',
                                                           'wasmer-compiler-cranelift-py310==1.0.0'],
 ':python_version >= "3.8" and python_version < "3.10"': ['wasmer==1.0.0',
                                                          'wasmer-compiler-cranelift==1.0.0']}

entry_points = \
{'console_scripts': ['mflt-compact-log.compact = mflt_compact_log.compact:main',
                     'mflt-compact-log.log_fmt = '
                     'mflt_compact_log.log_fmt:main']}

setup_kwargs = {
    'name': 'mflt-compact-log',
    'version': '0.0.1',
    'description': 'Memfault Compact Log Library',
    'long_description': '# Memfault Compact Log Library\n\nThis library enables decoding Memfault-flavored compact logs. For background\ninformation on compact logs, see here:\n\nhttps://mflt.io/compact-logs\n\n## Usage\n\nSome brief usage information below. See the source for detailed usage.\n\n### Extracting compact log format strings from .elf\n\nTo extract the format strings from the symbol file:\n\n```python\nfrom mflt_compact_log.log_fmt import LogFormatElfSectionParser\n\nelf = "path to elf file"\n# parse the elf file\nmappings = LogFormatElfSectionParser.get_mapping_from_elf_file(elf)\n# \'mappings\' is a dictionary mapping log id to LogFormatInfo data\nprint(mappings)\n\n>>> {8: LogFormatInfo(filename=\'./main/console_example_main.c\', line=245, n_args=0, fmt=\'This is a compact log example\')}\n```\n\n### Decoding compact logs\n\nTo apply the format string to raw compact log data:\n\n```python\nfrom mflt_compact_log.compact import CompactLogDecoder\n\n# example format string; this could instead be retrieved from the elf file\nfmt = "An Integer Format String: %d"\n\n# compact log hex encoded raw data\ncompact_log = "820A0B"\n\n# decode the log!\ncompact_log_as_bytes = bytes.fromhex(compact_log)\nlog = CompactLogDecoder.from_cbor_array(fmt, compact_log_as_bytes)\nlog.decode()\n>>> \'An Integer Format String: 11\'\n```\n',
    'author': 'Memfault Inc',
    'author_email': 'hello@memfault.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/memfault/memfault-firmware-sdk',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)
